@@ -1,7 +1,9 @@
 import logging
 from uuid import uuid4
+from py2neo.bulk import create_nodes, merge_nodes
+import os
+import json
 
-from graphio.queries import nodes_create_unwind, nodes_merge_unwind
 from graphio.objects.helper import chunks, create_single_index, create_composite_index
 from graphio import defaults
 from graphio.objects.relationshipset import RelationshipSet
@@ -14,7 +16,7 @@ class NodeSet:
     Container for a set of Nodes with the same labels and the same properties that define uniqueness.
     """
 
-    def __init__(self, labels, merge_keys=None, batch_size=None):
+    def __init__(self, labels, merge_keys=None, batch_size=None, default_props=None):
         """
 
         :param labels: The labels for the nodes in this NodeSet.
@@ -26,6 +28,7 @@ class NodeSet:
         """
         self.labels = labels
         self.merge_keys = merge_keys
+        self.default_props = default_props
 
         self.combined = '_'.join(sorted(self.labels)) + '_' + '_'.join(sorted(self.merge_keys))
         self.uuid = str(uuid4())
@@ -37,6 +40,9 @@ class NodeSet:
 
         self.nodes = []
 
+    def __str__(self):
+        return f"<NodeSet ({self.labels}; {self.merge_keys})>"
+
     def add_node(self, properties):
         """
         Create a node in this NodeSet.
@@ -44,7 +50,12 @@ class NodeSet:
         :param properties: Node properties.
         :type properties: dict
         """
-        self.nodes.append(properties)
+        if self.default_props:
+            node_props = {**self.default_props, **properties}
+        else:
+            node_props = properties
+
+        self.nodes.append(node_props)
 
     def add_nodes(self, list_of_properties):
         for properties in list_of_properties:
@@ -71,28 +82,42 @@ class NodeSet:
         # add node if not found
         self.add_node(properties)
 
-    def item_iterator(self):
-        """
-        Generator function that yields the node properties for all nodes in this NodeSet.
-
-        This is used to create chunks of the nodes without iterating all nodes. This can be removes in future
-        when NodeSet and RelationshipSet fully support generators (instead of lists of nodes/relationships).
-        """
-        for node in self.nodes:
-            yield node
-
     def to_dict(self):
         """
         Create dictionary defining the nodeset.
         """
-        return {"labels":self.labels,"merge_keys":self.merge_keys,"nodes":self.nodes}
+        return {"labels":self.labels, "merge_keys":self.merge_keys, "nodes":self.nodes}
 
     @classmethod
-    def from_dict(cls,nodeset_dict,batch_size=None):
+    def from_dict(cls,nodeset_dict, batch_size=None):
         ns = cls(labels=nodeset_dict["labels"],merge_keys=nodeset_dict["merge_keys"])
         ns.add_nodes(nodeset_dict["nodes"])
         return ns
 
+    def object_file_name(self, suffix: str = None) -> str:
+        """
+        Create a unique name for this NodeSet that indicates content. Pass an optional suffix.
+        NOTE: suffix has to include the '.' for a filename!
+
+            `nodeset_Label_merge-key_uuid`
+
+        With suffix:
+
+            `nodeset_Label_merge-key_uuid.json`
+        """
+        basename = f"nodeset_{'_'.join(self.labels)}_{'_'.join(self.merge_keys)}_{self.uuid}"
+        if suffix:
+            basename += suffix
+        return basename
+
+    def serialize(self, target_dir: str):
+        """
+        Serialize NodeSet to a JSON file in a target directory.
+        """
+        path = os.path.join(target_dir, self.object_file_name(suffix='.json'))
+        with open(path, 'wt') as f:
+            json.dump(self.to_dict(), f, indent=4)
+
     def create(self, graph, batch_size=None):
         """
         Create all nodes from NodeSet.
@@ -104,46 +129,13 @@ class NodeSet:
 
         i = 1
         for batch in chunks(self.nodes, size=batch_size):
-            batch = list(batch)
+
             log.debug('Batch {}'.format(i))
 
-            query = nodes_create_unwind(self.labels)
-            log.debug(query)
-
-            result = graph.run(query, props=batch)
+            create_nodes(graph, batch, labels=self.labels)
 
             i += 1
 
-    # TODO remove py2neo Node here, the node is just a dict now
-    def filter_nodes(self, filter_func):
-        """
-        Filter node properties with a filter function, remove nodes that do not match from main list.
-        """
-        filtered_nodes = []
-        discarded_nodes = []
-        for n in self.nodes:
-            node_properties = dict(n)
-            if filter_func(node_properties):
-                filtered_nodes.append(n)
-            else:
-                discarded_nodes.append(n)
-
-        self.nodes = filtered_nodes
-        self.discarded_nodes = discarded_nodes
-
-    # TODO remove py2neo Node here, the node is just a dict now
-    def reduce_node_properties(self, *keep_props):
-        filtered_nodes = []
-        for n in self.nodes:
-            new_props = {}
-            for k, v in dict(n).items():
-                if k in keep_props:
-                    new_props[k] = v
-
-            filtered_nodes.append(Node(*self.labels, **new_props))
-
-        self.nodes = filtered_nodes
-
     def merge(self, graph, merge_properties=None, batch_size=None):
         """
         Merge nodes from NodeSet on merge properties.
@@ -160,42 +152,14 @@ class NodeSet:
 
         log.debug('Batch Size: {}'.format(batch_size))
 
-        query = nodes_merge_unwind(self.labels, merge_properties)
-        log.debug(query)
-
         i = 1
         for batch in chunks(self.node_properties(), size=batch_size):
-            batch = list(batch)
+
             log.debug('Batch {}'.format(i))
-            log.debug(batch[0])
+            merge_nodes(graph, batch, (tuple(self.labels), *merge_properties))
 
-            graph.run(query, props=batch)
             i += 1
 
-    def map_to_1(self, graph, target_labels, target_properties, rel_type=None):
-        """
-        Create relationships from all nodes in this NodeSet to 1 target node.
-
-        :param graph: The py2neo Graph
-        :param other_node: The target node.
-        :param rel_type: Relationship Type
-        """
-
-        if not rel_type:
-            rel_type = 'FROM_SET'
-
-        rels = RelationshipSet(rel_type, self.labels, target_labels, self.merge_keys, target_properties)
-
-        for node in self.nodes:
-            # get properties for merge_keys
-            node_properties = {}
-            for k in self.merge_keys:
-                node_properties[k] = node[k]
-
-            rels.add_relationship(node_properties, target_properties, {})
-
-        rels.create(graph)
-
     def node_properties(self):
         """
         Yield properties of the nodes in this set. Used for create function.
