@@ -1,23 +1,43 @@
 from uuid import uuid4
 import logging
 import json
+import os
+from py2neo.bulk import create_relationships, merge_relationships
 
-from graphio.objects.relationship import Relationship
 from graphio import defaults
-from graphio.queries import query_create_rels_unwind, query_merge_rels_unwind
-from graphio.queries.query_parameters import params_create_rels_unwind_from_objects
 from graphio.objects.helper import chunks, create_single_index, create_composite_index
 
 log = logging.getLogger(__name__)
 
 
+def tuplify_json_list(list_object: list) -> tuple:
+    """
+    JSON.dump() stores tuples as JSON lists. This function receives a list (with sub lists)
+    and creates a tuple of tuples from the list. The tuples are the preferred input type for py2neo.
+
+    E.g.
+
+        [[a, b], [c, d]] -> ((a, b), (c, d))
+
+    :param list_object: A list with sub-lists.
+    :return: A tuple version of the list object.
+    """
+    output = tuple()
+    for element in list_object:
+        if isinstance(element, list):
+            output = output + (tuple(element), )
+        else:
+            output = output + (element, )
+    return output
+
+
 class RelationshipSet:
     """
     Container for a set of Relationships with the same type of start and end nodes.
     """
 
     def __init__(self, rel_type, start_node_labels, end_node_labels, start_node_properties, end_node_properties,
-                 batch_size=None):
+                 batch_size=None, default_props=None):
         """
 
         :param rel_type: Realtionship type.
@@ -39,6 +59,10 @@ class RelationshipSet:
         self.end_node_labels = end_node_labels
         self.start_node_properties = start_node_properties
         self.end_node_properties = end_node_properties
+        self.default_props = default_props
+
+        self.fixed_order_start_node_properties = tuple(self.start_node_properties)
+        self.fixed_order_end_node_properties = tuple(self.end_node_properties)
 
         self.uuid = str(uuid4())
         self.combined = '{0}_{1}_{2}_{3}_{4}'.format(self.rel_type,
@@ -58,122 +82,116 @@ class RelationshipSet:
         self.unique = False
         self.unique_rels = set()
 
+    def __str__(self):
+        return f"<RelationshipSet ({self.start_node_labels}; {self.start_node_properties})-[{self.rel_type}]->({self.end_node_labels}; {self.end_node_properties})>"
+
+    def __relationship_from_dictionary(self, start_node_properties, end_node_properties, properties):
+        """
+        Transform the input dictionary into the relationship data type used for py2neo.
+
+        :return: Relationship data.
+        """
+        start_node_data = tuple(start_node_properties[x] for x in self.fixed_order_start_node_properties)
+        if len(start_node_data) == 1:
+            start_node_data = start_node_data[0]
+        end_node_data = tuple(end_node_properties[x] for x in self.fixed_order_end_node_properties)
+        if len(end_node_data) == 1:
+            end_node_data = end_node_data[0]
+
+        return (start_node_data, properties, end_node_data)
+
     def add_relationship(self, start_node_properties, end_node_properties, properties):
         """
         Add a relationship to this RelationshipSet.
 
         :param properties: Relationship properties.
         """
+        if self.default_props:
+            rel_props = {**self.default_props, **properties}
+        else:
+            rel_props = properties
+
         if self.unique:
             # construct a check set with start_node_properties (values), end_node_properties (values) and properties (values)
             check_set = frozenset(
-                list(start_node_properties.values()) + list(end_node_properties.values()) + list(properties.values()))
+                list(start_node_properties.values()) + list(end_node_properties.values()) + list(rel_props.values()))
 
             if check_set not in self.unique_rels:
-                rel = Relationship(self.start_node_labels, self.end_node_labels, start_node_properties,
-                                   end_node_properties, properties)
-                self.relationships.append(rel)
+                self.relationships.append(self.__relationship_from_dictionary(start_node_properties, end_node_properties, rel_props))
                 self.unique_rels.add(check_set)
         else:
-            rel = Relationship(self.start_node_labels, self.end_node_labels, start_node_properties,
-                               end_node_properties, properties)
-            self.relationships.append(rel)
-
-    def item_iterator(self):
-        """
-        Generator function that yields the to_dict function for all relationships in this RelationshipSet.
-
-        This is used to create chunks of the relationships without iterating all relationships. This can be removes in future
-        when NodeSet and RelationshipSet fully support generators (instead of lists of nodes/relationships).
-        """
-        for rel in self.relationships:
-            yield rel.to_dict()
+            self.relationships.append(self.__relationship_from_dictionary(start_node_properties, end_node_properties, rel_props))
 
     def to_dict(self):
-        return {"rel_type":self.rel_type,
-                "start_node_labels":self.start_node_labels,
-                "end_node_labels":self.end_node_labels,
-                "start_node_properties":self.start_node_properties,
-                "end_node_properties":self.end_node_properties,
-                "unique":self.unique,
-                "relationships":[rel.to_dict() for rel in self.relationships]}
+        return {"rel_type": self.rel_type,
+                "start_node_labels": self.start_node_labels,
+                "end_node_labels": self.end_node_labels,
+                "start_node_properties": self.start_node_properties,
+                "end_node_properties": self.end_node_properties,
+                "unique": self.unique,
+                "relationships": self.relationships}
 
-    @classmethod  
-    def from_dict(cls,relationship_dict,batch_size=None):
+    @classmethod
+    def from_dict(cls, relationship_dict, batch_size=None):
         rs = cls(rel_type=relationship_dict["rel_type"],
-            start_node_labels=relationship_dict["start_node_labels"],
-            end_node_labels=relationship_dict["end_node_labels"],
-            start_node_properties=relationship_dict["start_node_properties"],
-            end_node_properties=relationship_dict["end_node_properties"],
-            batch_size=batch_size)
+                 start_node_labels=relationship_dict["start_node_labels"],
+                 end_node_labels=relationship_dict["end_node_labels"],
+                 start_node_properties=relationship_dict["start_node_properties"],
+                 end_node_properties=relationship_dict["end_node_properties"],
+                 batch_size=batch_size)
         rs.unique = relationship_dict["unique"]
-        [rs.add_relationship(start_node_properties=rel["start_node_properties"],end_node_properties=rel["end_node_properties"],properties=rel["properties"]) for rel in relationship_dict["relationships"]]
+        rs.relationships = [tuplify_json_list(r) for r in relationship_dict["relationships"]]
+
         return rs
 
-    def filter_relationships_target_node(self, filter_func):
+    def object_file_name(self, suffix: str = None) -> str:
         """
-        Filter properties of target node with a filter function, remove relationships that do not match from main list.
+        Create a unique name for this RelationshipSet that indicates content. Pass an optional suffix.
+        NOTE: suffix has to include the '.' for a filename!
+
+            `relationshipset_StartLabel_TYPE_EndLabel_uuid`
+
+        With suffix:
+
+            `relationshipset_StartLabel_TYPE_EndLabel_uuid.json`
         """
-        filtered_rels = []
-        discarded_rels = []
+        basename = f"relationshipset_{'_'.join(self.start_node_labels)}_{self.rel_type}_{'_'.join(self.end_node_labels)}_{self.uuid}"
+        if suffix:
+            basename += suffix
+        return basename
 
-        for rel in self.relationships:
-
-            if filter_func(rel.end_node_properties):
-                filtered_rels.append(rel)
-            else:
-                discarded_rels.append(rel)
-
-        self.relationships = filtered_rels
-        self.discarded_nodes = discarded_rels
-
-    def filter_relationships_start_node(self, filter_func):
+    def serialize(self, target_dir: str):
         """
-        Filter properties of target node with a filter function, remove relationships that do not match from main list.
+        Serialize NodeSet to a JSON file in a target directory.
         """
-        filtered_rels = []
-        discarded_rels = []
-
-        for rel in self.relationships:
-
-            if filter_func(rel.start_node_properties):
-                filtered_rels.append(rel)
-            else:
-                discarded_rels.append(rel)
-
-        self.relationships = filtered_rels
-        self.discarded_nodes = discarded_rels
-
-    def check_if_rel_exists(self, start_node_properties, end_node_properties, properties):
-        for rel in self.relationships:
-            if rel.start_node_properties == start_node_properties and rel.end_node_properties == end_node_properties and rel.properties == properties:
-                return True
+        path = os.path.join(target_dir, self.object_file_name(suffix='.json'))
+        with open(path, 'wt') as f:
+            json.dump(self.to_dict(), f, indent=4)
 
     def create(self, graph, batch_size=None):
         """
         Create relationships in this RelationshipSet
+
+        py2neo bulk works with tuples and th order of elements in the tuple. The underlying Relationship used in the
+        RelationshipSet uses a dictionary. Work around this for now, adapt the RelSet.add_relationship() method later.
         """
         log.debug('Create RelationshipSet')
         if not batch_size:
             batch_size = self.batch_size
         log.debug('Batch Size: {}'.format(batch_size))
 
-        # get query
-        query = query_create_rels_unwind(self.start_node_labels, self.end_node_labels, self.start_node_properties,
-                                         self.end_node_properties, self.rel_type)
-        log.debug(query)
-
         i = 1
         # iterate over chunks of rels
         for batch in chunks(self.relationships, size=batch_size):
-            batch = list(batch)
-            log.debug('Batch {}'.format(i))
-            log.debug(batch[0])
-            # get parameters
-            query_parameters = params_create_rels_unwind_from_objects(batch)
-            log.debug(json.dumps(query_parameters))
 
-            graph.run(query, **query_parameters)
+            log.debug('Batch {}'.format(i))
+
+            create_relationships(graph.auto(),
+                                 batch,
+                                 self.rel_type,
+                                 start_node_key=(tuple(self.start_node_labels), *self.fixed_order_start_node_properties),
+                                 end_node_key=(tuple(self.end_node_labels), *self.fixed_order_end_node_properties))
+
             i += 1
 
     def merge(self, graph, batch_size=None):
@@ -185,22 +203,18 @@ class RelationshipSet:
             batch_size = self.batch_size
         log.debug('Batch Size: {}'.format(batch_size))
 
-        # get query
-        query = query_merge_rels_unwind(self.start_node_labels, self.end_node_labels, self.start_node_properties,
-                                        self.end_node_properties, self.rel_type)
-        log.debug(query)
-
         i = 1
         # iterate over chunks of rels
         for batch in chunks(self.relationships, size=batch_size):
-            batch = list(batch)
             log.debug('Batch {}'.format(i))
-            log.debug(batch[0])
-            # get parameters
-            query_parameters = params_create_rels_unwind_from_objects(batch)
-            log.debug(json.dumps(query_parameters))
 
-            graph.run(query, **query_parameters)
+            merge_relationships(graph.auto(),
+                                batch,
+                                self.rel_type,
+                                start_node_key=(tuple(self.start_node_labels), *self.fixed_order_start_node_properties),
+                                end_node_key=(tuple(self.end_node_labels), *self.fixed_order_end_node_properties)
+                                )
+
             i += 1
 
     def create_index(self, graph):
@@ -227,7 +241,7 @@ class RelationshipSet:
         for label in self.end_node_labels:
             for prop in self.end_node_properties:
                 create_single_index(graph, label, prop)
-                
+
             # composite indexes
             if len(self.end_node_properties) > 1:
                 create_composite_index(graph, label, self.end_node_properties)
